/*
 *  NFSv4 ACL Code
 *  Print the contents of an nfs4 ACE
 *
 *  Permission mapping:
 *  r - NFS4_ACE_READ_DATA
 *  l - NFS4_ACE_LIST_DIRECTORY
 *  w - NFS4_ACE_WRITE_DATA
 *  f - NFS4_ACE_ADD_FILE
 *  a - NFS4_ACE_APPEND_DATA
 *  s - NFS4_ACE_ADD_SUBDIRECTORY
 *  n - NFS4_ACE_READ_NAMED_ATTRS
 *  N - NFS4_ACE_WRITE_NAMED_ATTRS
 *  x - NFS4_ACE_EXECUTE
 *  D - NFS4_ACE_DELETE_CHILD
 *  t - NFS4_ACE_READ_ATTRIBUTES
 *  T - NFS4_ACE_WRITE_ATTRIBUTES
 *  d - NFS4_ACE_DELETE
 *  c - NFS4_ACE_READ_ACL
 *  C - NFS4_ACE_WRITE_ACL
 *  o - NFS4_ACE_WRITE_OWNER
 *  y - NFS4_ACE_SYNCHRONIZE
 *
 *
 *  Copyright (c) 2005, 2006 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Andy Adamson <andros@umich.edu>
 *  David M. Richter <richterd@citi.umich.edu>
 *  Alexis Mackenzie <allamack@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <ctype.h>
#include "libacl_nfs4.h"

/*
 * An array of pointers to hold the parsed acetype, aceflag, acewho and
 * acemask.
 */
#define NUMFIELDS 	4
#define TYPE_INDEX	0
#define FLAG_INDEX	1
#define WHO_INDEX	2
#define MASK_INDEX	3


extern struct nfs4_ace * acl_nfs4_create_ace(int is_directory, u32 type, u32 flag, u32 access_mask,
			int whotype, char* who);

/*
 * strrep - replace all occurrences of 'from' with 'to' in the null-
 *  terminated string 's'
 */
void
strrep(char *s, char from, char to)
{
	char *t = s;
	
	while ((t = strchr(t, from)) != NULL)
		*t = to;
}

void
free_fields(char *fields[NUMFIELDS])
{
	int i;

	for (i = 0; i < NUMFIELDS; i++)
		if (fields[i] != NULL)
			free(fields[i]);
}

int
parse_alloc_fields(char *buf, char *fields[NUMFIELDS])
{
	char *prev, *next;
	int len, i,j;

	if (!buf)
		return -EINVAL;

	memset(fields, 0, sizeof(fields));
	prev = buf;
	for( i = 0; i < 3; i++) {
		next = strchr(prev, ':');
		if (next == NULL)
			return -EINVAL;
		len = next - prev;
		for(j = 0; j < len; j++) {
			if (prev[0] == ' ') {
				prev++;
				len--;
			}
		}
		fields[i] = malloc(len + 1);
		if (!fields[i])
			goto out_free;
		if (len > 0)
			memcpy(fields[i], prev, len);
		fields[i][len] = 0;
		prev = ++next;
	}
	len = strlen(prev);
	fields[NUMFIELDS - 1] = malloc(len + 1);
	if (len > 0)
		memcpy(fields[NUMFIELDS - 1], prev, len);
	fields[NUMFIELDS - 1][len] = 0;

	return 0;
out_free:
	free_fields(fields);
	return -ENOMEM;
}

/*
 * returns pointer to nfs4_ace on success, NULL on failure
 */

struct nfs4_ace * __acl_nfs4_string2ace(char *ace_buf,int is_dir)
{
	int ret;
	char *fields[NUMFIELDS];
	u32 type, flags = 0, mask = 0;
	char *bufp;
	int buflen;
	struct nfs4_ace * n_ace = NULL;

	strrep(ace_buf, '\n', '\0');

	/* e.g., we got a blank line or a comment */
	if (*ace_buf == '\0' || *ace_buf == '#')
		return NULL;
	
	/* remove ace index number, if present */
	bufp =  strchr(ace_buf, ':');
	if (isdigit(*(bufp - 1)))
        	bufp++;			/* we saw "<digit>: "; skip the space */
	else
		bufp = ace_buf;		/* ace isn't numbered */
	
	ret = parse_alloc_fields(bufp, fields);
	if (ret < 0) {
		printf("Scanning ACE string %s failed\n", bufp);
		goto out_fail;
	}

	switch (*fields[TYPE_INDEX]) {
		case 'A':
			type = NFS4_ACE_ACCESS_ALLOWED_ACE_TYPE;
			break;
		case 'D':
			type = NFS4_ACE_ACCESS_DENIED_ACE_TYPE;
			break;
		case 'U':
			type = NFS4_ACE_SYSTEM_AUDIT_ACE_TYPE;
			break;
		case 'L':
			type = NFS4_ACE_SYSTEM_ALARM_ACE_TYPE;
			break;
		default:
			printf("Bad Ace Type:%c\n", *fields[TYPE_INDEX]);
                        goto out_fail;
	}

	bufp = fields[FLAG_INDEX];
	for (buflen = strlen(fields[FLAG_INDEX]); buflen > 0; buflen--) {
		switch(*bufp) {
			case 'f':
				flags |= NFS4_ACE_FILE_INHERIT_ACE;
				break;
			case 'd':
				flags |= NFS4_ACE_DIRECTORY_INHERIT_ACE;
				break;
			case 'p':
				flags |= NFS4_ACE_NO_PROPAGATE_INHERIT_ACE;
				break;
			case 'i':
				flags |= NFS4_ACE_INHERIT_ONLY_ACE;
				break;
			case 'S':
				flags |= NFS4_ACE_SUCCESSFUL_ACCESS_ACE_FLAG;
				break;
			case 'F':
				flags |= NFS4_ACE_FAILED_ACCESS_ACE_FLAG;
				break;
			case 'g':
				flags |= NFS4_ACE_IDENTIFIER_GROUP;
				break;
			case 'O':
				flags |= NFS4_ACE_OWNER;
				break;
			case 'G':
				flags |= NFS4_ACE_GROUP;
				break;
			case 'E':
				flags |= NFS4_ACE_EVERYONE;
				break;
			default:
				printf("Bad Ace Type:%c\n", *bufp);
				goto out_fail;
		}
		bufp++;
	}

	bufp = fields[MASK_INDEX];
	for (buflen = strlen(fields[MASK_INDEX]); buflen > 0; buflen--) {
		ret = -EINVAL;
		switch(*bufp) {
			case 'l':
				if (!(is_dir & NFS4_ACL_ISDIR))
					goto out_not_dir;
				mask |= NFS4_ACE_LIST_DIRECTORY;
				break;
			case 'f':
				if (!(is_dir & NFS4_ACL_ISDIR))
					goto out_not_dir;
				mask |= NFS4_ACE_ADD_FILE;
				break;
			case 's':
				if (!(is_dir & NFS4_ACL_ISDIR))
					goto out_not_dir;
				mask |= NFS4_ACE_ADD_SUBDIRECTORY;
				break;
			case 'D':
				if (!(is_dir & NFS4_ACL_ISDIR))
					goto out_not_dir;
				mask |= NFS4_ACE_DELETE_CHILD;
				break;
			case 'r':
				if ((is_dir & NFS4_ACL_ISFILE))
					goto out_not_file;
				mask |= NFS4_ACE_READ_DATA;
				break;
			case 'w':
				if ((is_dir & NFS4_ACL_ISFILE))
					goto out_not_file;
				mask |= NFS4_ACE_WRITE_DATA;
				break;
			case 'a':
				if ((is_dir & NFS4_ACL_ISFILE))
					goto out_not_file;
				mask |= NFS4_ACE_APPEND_DATA;
				break;
			case 'n':
				mask |= NFS4_ACE_READ_NAMED_ATTRS;
				break;
			case 'N':
				mask |= NFS4_ACE_WRITE_NAMED_ATTRS;
				break;
			case 'x':
				mask |= NFS4_ACE_EXECUTE;
				break;
			case 't':
				mask |= NFS4_ACE_READ_ATTRIBUTES;
				break;
			case 'T':
				mask |= NFS4_ACE_WRITE_ATTRIBUTES;
				break;
			case 'd':
				mask |= NFS4_ACE_DELETE;
				break;
			case 'c':
				mask |= NFS4_ACE_READ_ACL;
				break;
			case 'C':
				mask |= NFS4_ACE_WRITE_ACL;
				break;
			case 'o':
				mask |= NFS4_ACE_WRITE_OWNER;
				break;
			case 'y':
				mask |= NFS4_ACE_SYNCHRONIZE;
				break;
			default:
				printf("Bad Ace Mask:%c\n", *bufp);
				goto out_fail;
		}
		bufp++;
	}

	n_ace = acl_nfs4_create_ace(is_dir, type, flags, mask,
				acl_nfs4_get_whotype(fields[WHO_INDEX]), fields[WHO_INDEX]);
	if (n_ace == NULL) {
		goto out_fail;
	}

out_fail:
	free_fields(fields);
	return n_ace;

out_not_dir:
		printf("Bad Ace Mask Directory %c mask on file\n", *bufp);
		goto out_fail;
out_not_file:
		printf("Bad Ace Mask File %c mask on Directory\n", *bufp);
		goto out_fail;
}
