/*  Copyright (c) 2002, 2003, 2006 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Andy Adamson <andros@citi.umich.edu>
 *  David M. Richter <richterd@citi.umich.edu>
 *  Alexis Mackenzie <allamack@citi.umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <attr/xattr.h>

#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "libacl_nfs4.h"

/* Actions */
#define NO_ACTION		0
#define MODIFY_ACTION		1
#define SUBSTITUTE_ACTION	2
#define REMOVE_ACTION		3
#define APPEND_ACTION		4

/* Walks */
#define DEFAULT_WALK		0	/* Follow symbolic link args, Skip links in subdirectories */
#define LOGICAL_WALK		1	/* Follow all symbolic links */
#define PHYSICAL_WALK		2	/* Skip all symbolic links */

/* Recursion */
#define NO_RECURSIVE		0
#define YES_RECURSIVE		1

/* Translation formats */
#define NO_TRANS		0	/* Straight NFSv4 ACL rules */
#define POSIX_LINUX_TRANS	1	/* Use NFSv4 POSIX Linux format */

#define u32 u_int32_t

typedef struct mod_ace_s mod_ace;
struct mod_ace_s {
	mod_ace* next;
	int cvals;
	char* user;
	u32 flags;
	u32 permissions;
	int line;
};

void usage(void);


/* XXX: things we need to handle:
 *
 *  - 'modify' operations should occur in-place if the ACEs are already in the ACL.
 *  - we need some sort of 'purge' operation that completely clears an ACL.
 *  - how to handle 'remove'?  should they have to enter full ACEs, or maybe just the principal's
 *  	name?  or maybe purge could even optionally take a principal's name?  or maybe 'purge' and
 *  	'remove' could be the same operation if we can resolve what the default behaviors should
 *  	be..
 */

int main(int argc, char** argv)
{
	int opt, res;
	int trans_type = NO_TRANS;
	int do_recursive = NO_RECURSIVE;
	int walk_type = DEFAULT_WALK;
	int action = NO_ACTION;
	FILE *s_fd;
	char* spec_file = NULL;
	char* mod_string = NULL;
	struct nfs4_acl *nacl;
	struct stat st;
	char *xdrbuf;
	char * path_p = argv[argc-1];

	while(1)
	{
		opt = getopt(argc, argv, "jLa:A:m:s:M:PRS:x:X:");

		if (opt == -1) {
			if (optind >= argc) {
				usage();
				goto failed;
			}
			break;
		}
		else if (opt == ':') {
			usage();
			goto failed;
		}
		else if (opt == '?') {
			usage();
			goto failed;
		}


		/* We have a good option */
		switch(opt) {
			case 'j':
				trans_type = POSIX_LINUX_TRANS;
				break;

			case 'L':
				if (walk_type != DEFAULT_WALK) {
					fprintf(stderr, "More than one Walk type specified\n");
					usage();
					goto failed;
				}
				walk_type = LOGICAL_WALK;
				break;
			case 'a':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = APPEND_ACTION;
				mod_string = optarg;
				break;
			case 'A':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = APPEND_ACTION;
				spec_file = optarg;
				s_fd = fopen (spec_file, "r");
				if (s_fd == NULL) {
					fprintf(stderr, "Access error on spec file: %s\n", spec_file);
					goto failed;
				}
				break;


			case 'm':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = MODIFY_ACTION;
				mod_string = optarg;
				break;

			case 'M':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = MODIFY_ACTION;
				spec_file = optarg;
				s_fd = fopen (spec_file, "r");
				if (s_fd == NULL) {
					fprintf(stderr, "Access error on spec file: %s\n", spec_file);
					goto failed;
				}
				break;

			case 'P':
				if (walk_type != DEFAULT_WALK) {
					fprintf(stderr, "More than one walk type specified\n");
					usage();
					goto failed;
				}
				walk_type = PHYSICAL_WALK;
				break;

			case 'R':
				do_recursive = YES_RECURSIVE;
				break;
			case 's':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = SUBSTITUTE_ACTION;
				mod_string = optarg;
				break;

			case 'S':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = SUBSTITUTE_ACTION;
				spec_file = optarg;
				s_fd = fopen (spec_file, "r");
				if (s_fd == NULL) {
					fprintf(stderr, "Access error on spec file: %s \n", spec_file);
					goto failed;
				}
				break;

			case 'x':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = REMOVE_ACTION;
				mod_string = optarg;
				break;

			case 'X':
				if (action != NO_ACTION) {
					fprintf(stderr, "More than one action specified\n");
					usage();
					goto failed;
				}
				action = REMOVE_ACTION;
				spec_file = optarg;
				s_fd = fopen (spec_file, "r");
				if (s_fd == NULL) {
					fprintf(stderr, "Access error on spec file: %s\n", spec_file);
					goto failed;
				}
				break;
		}
	}

	if (action == NO_ACTION) {
		fprintf(stderr, "No action specified\n");
		usage();
		goto failed;
	}

	res = stat(path_p, &st);
	if (res < 0) {
		perror(path_p);
		goto failed;
	}

	if ((nacl = acl_nfs4_new(st.st_mode & S_IFDIR ? 1 : 0)) == NULL) {
		fprintf(stderr, "Failed to allocate memory\n");
		errno = ENOMEM;
		goto failed;
	}

	
	if (action != SUBSTITUTE_ACTION) 
		nacl = nfs4_acl_for_path(path_p);
	
	if (action == APPEND_ACTION) {
		//maybe name of 'nfs4_init_acl*' should be changed seem as it also does the append
		//NOTE:total_who_length isn't going to be correct if acl already contains entries
		if (mod_string) {
			if (nfs4_init_acl_from_string(mod_string, &nacl) < 0) {
				fprintf(stderr, "Failed while building ACL from string \"%s\".  Aborting.\n", mod_string);
				goto failed_free;
			}
		} else {		
			if ((s_fd != NULL) && (res = nfs4_init_acl_from_file(s_fd, &nacl) < 0)) {
				fprintf(stderr, "Failed loading ACL from spec file %s. Aborting.\n", spec_file);
				goto failed_free;
			} else if (s_fd == NULL) {
				fprintf(stderr, "Partial Implementation: No spec file specified, Aborting.\n");
				goto failed_free;
			}
		}
	} else if (action == REMOVE_ACTION) {
		if (mod_string) {
			if (nfs4_remove_string_entries_from_acl(mod_string,&nacl) < 0) {
				fprintf(stderr, "Cannot remove entries from an empty acl. Aborting.\n");
				goto failed_free;
			}
		} else {
			if (nfs4_remove_file_entries_from_acl(s_fd,&nacl) < 0) {
				fprintf(stderr, "Cannot remove entries from an empty acl. Aborting.\n");
				goto failed_free;
			}
		}
	} else if (action == MODIFY_ACTION) { //needs error checking
		if (mod_string) {
			if (nfs4_replace_acl_entry_from_string(mod_string,&nacl) < 0) {
				fprintf(stderr, "Failed while trying to replace acl entry. Aborting\n");
				goto failed_free;
			}
		} else {
			if (nfs4_replace_acl_entry_from_file(s_fd,&nacl) < 0) {
				fprintf(stderr, "Failed while trying to replace acl entry. Aborting\n");
				goto failed_free;
			}
		}
	} else if (action == SUBSTITUTE_ACTION) {
		if (mod_string) {
			if (nfs4_init_acl_from_string(mod_string, &nacl) < 0) {
				fprintf(stderr, "Failed while building ACL from string \"%s\".  Aborting.\n", mod_string);
				goto failed_free;
			}

		} else {
			if ((s_fd != NULL) && (res = nfs4_init_acl_from_file(s_fd, &nacl) < 0)) {
				fprintf(stderr, "Failed loading ACL from spec file %s. Aborting.\n", spec_file);
				goto failed_free;
			} else if (s_fd == NULL) {
				fprintf(stderr, "Partial Implementation: No spec file specified, Aborting.\n");
				goto failed_free;
			}
		}
	}

	res = nfs4_set_acl(nacl, path_p);
	acl_nfs4_free(nacl);
	return res;

failed:
	return 1;

failed_free:
	acl_nfs4_free(nacl);
	goto failed;
}



void usage(void)
{
	fprintf(stderr, "usage (under development): nfs4_setfacl [-j] [-M spec_file] | [-m acl_spec]  file\n");
	fprintf(stderr, "  -S spec_file	 set NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -s acl_spec	 set NFSv4 ACL with acl_spec string\n");
	fprintf(stderr, "  -A spec_file	 append contents of spec_file onto existing NFSv4 ACL\n");
	fprintf(stderr, "  -a acl_spec	 append acl_spec string onto existing NFSv4 ACL\n");
	fprintf(stderr, "  -M spec_file	 modify NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -m acl_spec	 modify NFSv4 ACL with contents of acl_spec\n");
	fprintf(stderr, "  -X spe_file	 read ACL entries to remove from spec file\n");
	fprintf(stderr, "  -x acl_spec	 remove entries from the ACL(s) of file(s)\n");
}

#if 0
void usage(void)
{
	fprintf(stderr, "usage: nfs4_setfacl [-jRLP] [{-m|-x} acl_spec] [{-M|-X} acl_file] file ...\n");
	fprintf(stderr, "  -S spec_file	 set NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -M spec_file	 modify NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -m acl_spec	 modify NFSv4 ACL with contents of acl_spec\n");
	fprintf(stderr, "  -s acl_spec	 set NFSv4 ACL with acl_spec string\n");
	fprintf(stderr, "  -a acl_spec   append NFSv4 ACL entry onto existing acl\n");
	fprintf(stderr, "  -A spec_file	 append contents of spec_file onto existing NFSv4 ACL\n");
	fprintf(stderr, "  -x acl_spec	 remove entries from the ACL(s) of file(s)\n");
	fprintf(stderr, "  -X spe_file	 read ACL entries to remove from spec file\n");
	fprintf(stderr, "  -j		 format ACL to LINUX POSIX translation spec\n");
	fprintf(stderr, "  -R		 recurse into subdirectories\n");
	fprintf(stderr, "  -L		 logical walk, follow symbolic links\n");
	fprintf(stderr, "  -P		 physical walk, do not follow symbolic links\n");
	return;
}
#endif
